#include <string.h>

#include <openssl/evp.h>

#include "stun.h"

static EVP_AEAD_CTX ectx, dctx;
static const EVP_AEAD *aead;
static unsigned char key[EVP_MAX_KEY_LENGTH];

static struct cipher {
	const char *name;
	const EVP_AEAD *(*aeadfn)(void);
} ciphers[] = {
	{ "aes-128-gcm",            EVP_aead_aes_128_gcm            },
	{ "aes-256-gcm",            EVP_aead_aes_256_gcm            },
	{ "chacha20-poly1305",      EVP_aead_chacha20_poly1305      },
#if LIBRESSL_VERSION_NUMBER >= 0x2030200fL
	{ "chacha20-poly1305-ietf", EVP_aead_chacha20_poly1305_ietf },
#endif
	{ NULL, NULL }
};

void
listciphers(void)
{
	struct cipher *cp;

	for (cp = ciphers; cp->name; cp++)
		puts(cp->name);
}

void
setcipher(char *name)
{
	struct cipher *cp;

	for (cp = ciphers; cp->name; cp++) {
		if (strcmp(cp->name, name) == 0) {
			aead = (*cp->aeadfn)();
			return;
		}
	}
	logerr("unknown cipher: %s", name);
}

void
derivekey(char *pw)
{
	if (!PKCS5_PBKDF2_HMAC_SHA1(pw, strlen(pw), NULL, 0, NROUNDS,
		                    EVP_AEAD_key_length(aead), key))
		logerr("PKCS5_PBKDF2_HMAC_SHA1 failed");
}

void
cryptoinit(void)
{
	if (!EVP_AEAD_CTX_init(&ectx, aead, key, EVP_AEAD_key_length(aead),
	                       EVP_AEAD_DEFAULT_TAG_LENGTH, NULL) ||
	    !EVP_AEAD_CTX_init(&dctx, aead, key, EVP_AEAD_key_length(aead),
	                       EVP_AEAD_DEFAULT_TAG_LENGTH, NULL))
		logerr("EVP_AEAD_CTX_init failed");
}

size_t
cryptononcelen(void)
{
	return EVP_AEAD_nonce_length(aead);
}

size_t
cryptotaglen(void)
{
	return EVP_AEAD_max_tag_len(aead);
}

int
cryptoseal(unsigned char *out, size_t *outlen, size_t maxoutlen,
           const unsigned char *nonce, size_t noncelen, const unsigned char *in,
           size_t inlen, const unsigned char *ad, size_t adlen)
{
	return EVP_AEAD_CTX_seal(&ectx, out, outlen, maxoutlen, nonce, noncelen,
	                         in, inlen, ad, adlen);
}

int
cryptoopen(unsigned char *out, size_t *outlen, size_t maxoutlen,
           const unsigned char *nonce, size_t noncelen, const unsigned char *in,
           size_t inlen, const unsigned char *ad, size_t adlen)
{
	return EVP_AEAD_CTX_open(&dctx, out, outlen, maxoutlen, nonce, noncelen,
	                         in, inlen, ad, adlen);
}
