#include <errno.h>
#include <stdint.h>
#include <stdlib.h>
#include <unistd.h>

#if defined(__linux__)
#include <bsd/stdlib.h>
#endif

#include "stun.h"

/* the various states the input handler can be in */
enum {
	INPUTINITIAL,
	INPUTNONCE,
	INPUTHDR,
	INPUTPAYLOAD,
	INPUTTAG,
	INPUTDECRYPT,
	INPUTDISCARD
};

/* output buffer for writing data to the network */
static unsigned char *wbuf;
/* input buffer for reading data from the network */
static unsigned char *rbuf;

/* total number of bytes read into input buffer */
static size_t rbuftotal;
/* number of bytes remaining to read into buffer for current state */
static size_t rbufrem;

/* maximum input/output buffer size */
static size_t maxbuflen;

/* nonce size in bytes */
static size_t noncelen;
/* tag size in bytes */
static size_t taglen;

/* state tracking for input handling */
static int state;

/* create a packet and write it out to the network */
int
netwrite(int fd, unsigned char *pt, size_t ptlen, size_t *outlen)
{
	unsigned char *p = wbuf;
	size_t buflen = noncelen + HDRLEN + ptlen + taglen;
	int n, total = 0;

	if (buflen > maxbuflen)
		return PKTFAILED;

	arc4random_buf(wbuf, noncelen);
	pack16(&wbuf[noncelen], ptlen);
	if (!cryptoseal(&wbuf[noncelen + HDRLEN], outlen,
	                ptlen + taglen, wbuf, noncelen,
	                pt, ptlen, &wbuf[noncelen], HDRLEN)) {
		logwarn("cryptoseal failed");
		return -1;
	}
	*outlen = ptlen;

	/* spin until all data is written */
	while (buflen > 0) {
		n = write(fd, p + total, buflen);
		if (n == 0) {
			return PKTFAILED;
		} else if (n < 0) {
			if (errno == EWOULDBLOCK)
				continue;
			return PKTFAILED;
		}
		total += n;
		buflen -= n;
	}
	return PKTCOMPLETE;
}

/*
 * Since stun is using non-blocking sockets it needs to handle
 * partial reads.  A state machine is devised for this purpose.
 */
int
netread(int fd, unsigned char *pt, size_t ptlen, size_t *outlen)
{
	size_t buflen = noncelen + HDRLEN + ptlen + taglen;
	int n, ctlen;

	if (buflen > maxbuflen)
		return PKTFAILED;

	for (;;) {
		switch (state) {
		case INPUTINITIAL:
			rbuftotal = 0;
			rbufrem = noncelen;
			state = INPUTNONCE;
			break;
		case INPUTNONCE:
			if ((n = read(fd, rbuf + rbuftotal, rbufrem)) <= 0)
				goto out;
			rbuftotal += n;
			rbufrem -= n;
			if (rbufrem == 0) {
				rbufrem = HDRLEN;
				state = INPUTHDR;
			}
			break;
		case INPUTHDR:
			if ((n = read(fd, rbuf + rbuftotal, rbufrem)) <= 0)
				goto out;
			rbuftotal += n;
			rbufrem -= n;
			if (rbufrem == 0) {
				ctlen = unpack16(&rbuf[noncelen]);
				if (ctlen > ptlen) {
					state = INPUTDISCARD;
				} else {
					rbufrem = ctlen;
					state = INPUTPAYLOAD;
				}
			}
			break;
		case INPUTPAYLOAD:
			if ((n = read(fd, rbuf + rbuftotal, rbufrem)) <= 0)
				goto out;
			rbuftotal += n;
			rbufrem -= n;
			if (rbufrem == 0) {
				rbufrem = taglen;
				state = INPUTTAG;
			}
			break;
		case INPUTTAG:
			if ((n = read(fd, rbuf + rbuftotal, rbufrem)) <= 0)
				goto out;
			rbuftotal += n;
			rbufrem -= n;
			if (rbufrem == 0)
				state = INPUTDECRYPT;
			break;
		case INPUTDECRYPT:
			state = INPUTINITIAL;
			if (!cryptoopen(pt, outlen, ptlen, rbuf, noncelen,
			                &rbuf[noncelen + HDRLEN],
			                rbuftotal - noncelen - HDRLEN,
			                &rbuf[noncelen], HDRLEN)) {
				logwarn("cryptoopen failed");
				return PKTPARTIAL;
			}
			return PKTCOMPLETE;
		case INPUTDISCARD:
			state = INPUTINITIAL;
			for (;;)
				if ((n = read(fd, rbuf, maxbuflen)) <= 0)
					goto out;
			break;
		}
	}
out:
	if (n < 0 && errno == EWOULDBLOCK)
		return PKTPARTIAL;
	return PKTFAILED;
}

/* Reset state machine.  This is required when a fatal error occurs. */
void
netreset(void)
{
	state = INPUTINITIAL;
}

void
netinit(void)
{
	noncelen = cryptononcelen();
	taglen = cryptotaglen();
	maxbuflen = noncelen + HDRLEN + MAXPAYLOADLEN + taglen;
	if (!(wbuf = malloc(maxbuflen)))
		logerr("oom");
	if (!(rbuf = malloc(maxbuflen)))
		logerr("oom");
	netreset();
}
