/*
 * Design overview:
 *
 * stun implements a point to point encrypted tunnel.  It supports
 * layer 2 (TAP) and layer 3 (TUN) tunnels.  TCP is the only supported
 * transport.
 *
 * stun is a client-server design.  There can only be a single client
 * connected to the server at any time.  Routing between clients is handled
 * by the networking stack on the server side.
 *
 * stun supports multiple AEAD ciphers:
 *
 *      aes-128-gcm
 *      aes-256-gcm
 *      chacha20-poly1305
 *      chacha20-poly1305-ietf
 *
 * The default cipher is chacha20-poly1305.  The key is derived with
 * PBKDF.
 *
 * When a client connects there is a mutual challenge-response phase
 * as shown below:
 *
 * t0: server challenges client
 * t1: client responds to server's challenge
 * t2: client challenges server
 * t3: server responds to client's challenge
 *
 * The challenge-response algorithm is as follows:
 *
 * The challenge is a randomly generated 64-bit integer encrypted
 * with the pre-shared key and sent to the receiver.
 * The receiver decrypts it, adds 1, encrypts it and sends it back to
 * the sender.  The sender verifies the response.
 * This algorithm is the same as what Kerberos uses for authentication.
 *
 * The stun packet format is shown below:
 *
 * [NONCE] [PAYLOAD LEN] [PAYLOAD] [TAG]
 *
 */

#include <sys/resource.h>

#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "arg.h"
#include "stun.h"

char *argv0;
char *bindaddr;
char *host;
char *port = DEFPORT;
char *cipher = DEFCIPHER;
int devtype = TUNDEV;
int aftype = AF_INET;
int debug;
int sflag;

void
usage(void)
{
	fprintf(stderr, "usage: stun [-46] [-d] -s [-b address] [-p port] [-t devtype] [-c cipher] interface\n");
	fprintf(stderr, "       stun [-46] [-d] -h host [-p port] [-t devtype] [-c cipher] interface\n");
	exit(1);
}

int
main(int argc, char *argv[])
{
	struct rlimit rlim;
	char *arg, *pw;
	int devfd, listenfd, netfd;

	ARGBEGIN {
	case '4':
		aftype = AF_INET;
		break;
	case '6':
		aftype = AF_INET6;
		break;
	case 'd':
		debug = 1;
		break;
	case 's':
		sflag = 1;
		break;
	case 'b':
		bindaddr = EARGF(usage());
		break;
	case 'h':
		host = EARGF(usage());
		break;
	case 'p':
		port = EARGF(usage());
		break;
	case 't':
		arg = EARGF(usage());
		if (strcasecmp(arg, "tun") == 0)
			devtype = TUNDEV;
		else if (strcasecmp(arg, "tap") == 0)
			devtype = TAPDEV;
		else
			usage();
		break;
	case 'c':
		cipher = EARGF(usage());
		if (strcmp(cipher, "?") == 0) {
			listciphers();
			return 0;
		}
		break;
	default:
		usage();
	} ARGEND

	if (argc != 1 || !(sflag ^ (host != NULL)))
		usage();

	loginit("stun");

	/* disable core dumps as memory contains the pre-shared key */
	rlim.rlim_cur = rlim.rlim_max = 0;
	if (setrlimit(RLIMIT_CORE, &rlim) < 0)
		logerr("failed to disable core dumps");

	signal(SIGPIPE, SIG_IGN);
	if (!debug)
		daemon(0, 0);

	/* initialize tun/tap device */
	devfd = devopen(argv[0]);

	/* initialize crypto engine */
	if (!(pw = getenv("STUNPW")))
		logerr("STUNPW is not set");
	setcipher(cipher);
	derivekey(pw);
	cryptoinit();
	memset(pw, 0, strlen(pw));

	/* initialize networking engine */
	netinit();

	if (sflag) {
		/* invoked as server */
		listenfd = serverinit(bindaddr, port);
		revokeprivs();
#if defined(__OpenBSD__)
#include <sys/param.h>
#if OpenBSD >= 201605
		if (pledge("stdio inet", NULL) < 0)
			logerr("pledge failed");
#endif
#endif
		for (;;) {
			if ((netfd = serveraccept(listenfd)) < 0) {
				netreset();
				continue;
			}
			logdbg("client %s is ready", peer_ntop(netfd));
			tunnel(netfd, devfd);
			logdbg("client %s disconnected", peer_ntop(netfd));
			close(netfd);
			netreset();
		}
	} else {
		/* invoked as client */
		revokeprivs();
#if defined(__OpenBSD__)
#include <sys/param.h>
#if OpenBSD >= 201605
		if (pledge("stdio dns inet", NULL) < 0)
			logerr("pledge failed");
#endif
#endif
		for (;;) {
			if ((netfd = clientconnect(host, port)) < 0) {
				netreset();
				sleep(RECONNECTTIMEO);
				continue;
			}
			logdbg("connected to %s", peer_ntop(netfd));
			tunnel(netfd, devfd);
			logdbg("disconnected from %s", peer_ntop(netfd));
			close(netfd);
			netreset();
			sleep(RECONNECTTIMEO);
		}
	}
	return 0;
}
